from .parameter import parameter
from .pipeline import pipeline
from .run import current_run
from .runtime import download_pipeline, import_pipeline

__all__ = [
    "pipeline",
    "parameter",
    "current_run",
    "import_pipeline",
    "download_pipeline",
]



import pytest

from openhexa.sdk.pipelines.parameter import (
    Boolean,
    Float,
    FunctionWithParameter,
    Integer,
    InvalidParameterError,
    Parameter,
    ParameterValueError,
    String,
    parameter,
)


def test_parameter_types_normalize():
    # String
    string_parameter_type = String()
    assert string_parameter_type.normalize("a string") == "a string"
    assert string_parameter_type.normalize(" a string ") == "a string"
    assert string_parameter_type.normalize("") is None
    assert string_parameter_type.normalize(" ") is None

    # Integer
    integer_parameter_type = Integer()
    assert integer_parameter_type.normalize(99) == 99
    assert integer_parameter_type.normalize("abc") == "abc"

    # Float
    float_parameter_type = Float()
    assert float_parameter_type.normalize(3.14) == 3.14
    assert float_parameter_type.normalize(3) == 3.0

    # Boolean
    boolean_parameter_type = Boolean()
    assert boolean_parameter_type.normalize(True) is True
    assert boolean_parameter_type.normalize(False) is False
    assert boolean_parameter_type.normalize(3) == 3


def test_parameter_types_validate():
    # String
    string_parameter_type = String()
    assert string_parameter_type.validate("a string") == "a string"
    with pytest.raises(ParameterValueError):
        string_parameter_type.validate(86)

    # Integer
    integer_parameter_type = Integer()
    assert integer_parameter_type.validate(99) == 99
    with pytest.raises(ParameterValueError):
        integer_parameter_type.validate("not an int")

    # Float
    float_parameter_type = Float()
    assert float_parameter_type.validate(3.14) == 3.14
    with pytest.raises(ParameterValueError):
        float_parameter_type.validate("3.14")

    # Boolean
    boolean_parameter_type = Boolean()
    assert boolean_parameter_type.validate(True) is True
    assert boolean_parameter_type.validate(False) is False
    with pytest.raises(ParameterValueError):
        boolean_parameter_type.validate(86)


def test_parameter_init():
    # Wrong type
    with pytest.raises(InvalidParameterError):
        Parameter("arg", type="string")

    # Wrong code
    with pytest.raises(InvalidParameterError):
        Parameter("-123", type=str)
    with pytest.raises(InvalidParameterError):
        Parameter("Abc", type=str)
    with pytest.raises(InvalidParameterError):
        Parameter("0_z", type=str)

    # Empty choices
    with pytest.raises(InvalidParameterError):
        Parameter("arg", type=str, choices=[])

    # Invalid choices
    with pytest.raises(InvalidParameterError):
        Parameter("arg", type=str, choices=[1, 2, 3])

    # Boolean can't have choices
    with pytest.raises(InvalidParameterError):
        Parameter("arg", type=bool, choices=[True, False])

    # Boolean can't be multiple
    with pytest.raises(InvalidParameterError):
        Parameter("arg", type=bool, multiple=True)

    # Invalid defaults
    with pytest.raises(InvalidParameterError):
        Parameter("arg", type=bool, default=3)
    with pytest.raises(InvalidParameterError):
        Parameter("arg", type=str, default=[1, 2], multiple=True)
    with pytest.raises(InvalidParameterError):
        Parameter("arg", type=str, default="yolo", multiple=True)
    with pytest.raises(InvalidParameterError):
        Parameter("arg", type=str, default="")
    with pytest.raises(InvalidParameterError):
        Parameter("arg", type=str, default=[""], multiple=True)


def test_parameter_validate_single():
    # required is True by default
    parameter_1 = Parameter("arg1", type=str)
    assert parameter_1.validate("a valid string") == "a valid string"
    with pytest.raises(ParameterValueError):
        parameter_1.validate(None)
    with pytest.raises(ParameterValueError):
        parameter_1.validate("")

    # still required, but a default is provided
    parameter_2 = Parameter("arg2", type=int, default=3)
    assert parameter_2.validate(None) == 3

    # not required, no default
    parameter_3 = Parameter("arg3", type=bool, required=False)
    assert parameter_3.validate(None) is None

    # choices
    parameter_4 = Parameter("arg4", type=str, choices=["ab", "cd"])
    assert parameter_4.validate("ab") == "ab"
    with pytest.raises(ParameterValueError):
        parameter_4.validate("xy")


def test_parameter_validate_multiple():
    # required is True by default
    parameter_1 = Parameter("arg1", type=str, multiple=True)
    assert parameter_1.validate(["Valid string", "Another valid string"]) == [
        "Valid string",
        "Another valid string",
    ]
    with pytest.raises(ParameterValueError):
        parameter_1.validate(None)
    with pytest.raises(ParameterValueError):
        parameter_1.validate([])
    with pytest.raises(ParameterValueError):
        parameter_1.validate(["", ""])

    # still required, but a default is provided
    parameter_2 = Parameter("arg2", type=int, multiple=True, default=[2, 3])
    assert parameter_2.validate(None) == [2, 3]

    # not required, no default
    parameter_3 = Parameter("arg3", type=int, required=False, multiple=True)
    assert parameter_3.validate(None) == []
    assert parameter_3.validate([]) == []

    # choices
    parameter_4 = Parameter("arg4", type=str, choices=["ab", "cd"], multiple=True)
    assert parameter_4.validate(["ab"]) == ["ab"]
    assert parameter_4.validate(["ab", "cd"]) == ["ab", "cd"]
    with pytest.raises(ParameterValueError):
        parameter_4.validate(["xy"])
    with pytest.raises(ParameterValueError):
        parameter_4.validate(["ab", "xy"])


def test_parameter_parameters_spec():
    # required is True by default
    an_parameter = Parameter("arg1", type=str, default="yep")
    another_parameter = Parameter(
        "arg2",
        type=str,
        name="Arg 2",
        help="Help 2",
        choices=["ab", "cd"],
        required=False,
        multiple=True,
    )

    assert an_parameter.parameter_spec() == {
        "code": "arg1",
        "name": None,
        "type": "str",
        "required": True,
        "choices": None,
        "help": None,
        "multiple": False,
        "default": "yep",
    }

    assert another_parameter.parameter_spec() == {
        "code": "arg2",
        "name": "Arg 2",
        "type": "str",
        "required": False,
        "choices": ["ab", "cd"],
        "help": "Help 2",
        "multiple": True,
        "default": None,
    }


def test_parameter_decorator():
    @parameter("arg1", type=int)
    @parameter(
        "arg2",
        type=str,
        name="Arg 2",
        help="Help 2",
        default=["yo"],
        required=False,
        multiple=True,
    )
    def a_function():
        pass

    assert isinstance(a_function, FunctionWithParameter)
    function_parameters = a_function.get_all_parameters()
    assert len(function_parameters) == 2
    assert all(isinstance(a, Parameter) for a in function_parameters)

    assert function_parameters[0].code == "arg1"
    assert isinstance(function_parameters[0].type, Integer)
    assert function_parameters[0].name is None
    assert function_parameters[0].help is None
    assert function_parameters[0].default is None
    assert function_parameters[0].required is True
    assert function_parameters[0].multiple is False

    assert function_parameters[1].code == "arg2"
    assert isinstance(function_parameters[1].type, String)
    assert function_parameters[1].name == "Arg 2"
    assert function_parameters[1].help == "Help 2"
    assert function_parameters[1].default == ["yo"]
    assert function_parameters[1].required is False
    assert function_parameters[1].multiple is True



import sys
from pathlib import Path

import click
import stringcase

from openhexa import __version__
from openhexa.cli.api import (
    create_pipeline,
    ensure_is_pipeline_dir,
    get_pipeline,
    get_pipelines,
    get_workspace,
    is_debug,
    open_config,
    save_config,
    upload_pipeline,
)
from openhexa.sdk.pipelines import import_pipeline


@click.group()
@click.option("--debug/--no-debug", default=False, envvar="DEBUG")
@click.version_option(__version__)
@click.pass_context
def app(ctx, debug):
    """
    OpenHexa CLI
    """
    # ensure that ctx.obj exists and is a dict (in case `cli()` is called
    # by means other than the `if` block below)
    ctx.ensure_object(dict)

    user_config = open_config()
    user_config.set("openhexa", "debug", str(debug))
    save_config(user_config)


@app.group(invoke_without_command=True)
@click.pass_context
def workspaces(ctx):
    """
    Manage workspaces (add workspace, remove workspace, list workspaces, activate a workspace)
    """
    if ctx.invoked_subcommand is None:
        ctx.forward(workspaces_list)


@workspaces.command(name="add")
@click.argument("slug")
@click.option("--token", prompt=True, hide_input=True, confirmation_prompt=False)
def workspaces_add(slug, token):
    """
    Add a workspace to the configuration and activate it. The access token is required to access the workspace.
    """
    user_config = open_config()
    if slug in user_config["workspaces"]:
        click.echo(f"Workspace {slug} already exists. We will only update its token.")
    else:
        click.echo(f"Adding workspace {slug}")
    try:
        get_workspace(user_config, slug, token)
    except Exception as e:
        click.echo(
            f"Error while getting workspace '{slug}' on {user_config['openhexa']['url']}. Check the slug of the workspace and the access token.",
            err=True,
        )
        if is_debug(user_config):
            raise e
        return click.Abort()

    user_config["workspaces"].update({slug: token})
    user_config["openhexa"].update({"current_workspace": slug})

    save_config(user_config)


@workspaces.command(name="activate")
@click.argument("slug")
def workspaces_activate(slug):
    """
    Activate a workspace that is already in the configuration. The activated workspace will be used for the 'pipelines' commands.
    """

    user_config = open_config()
    if slug not in user_config["workspaces"]:
        click.echo(
            f"Workspace {slug} does not exist on {user_config['openhexa']['url']}. Available workspaces:"
        )
        click.echo(", ".join(user_config["workspaces"].keys()))
        return click.Abort()
    click.echo(f"Activating workspace {slug}")
    user_config["openhexa"].update({"current_workspace": slug})

    save_config(user_config)


@workspaces.command(name="list")
def workspaces_list():
    """
    List the workspaces in the configuration.
    """
    user_config = open_config()

    click.echo("Workspaces:")
    for slug in user_config["workspaces"]:
        click.echo(
            click.style(f"* {slug} (active)", bold=True)
            if slug == user_config["openhexa"]["current_workspace"]
            else f"* {slug}"
        )

    save_config(user_config)


@workspaces.command(name="rm")
@click.argument("slug")
def workspaces_rm(slug):
    """
    Remove a workspace from the configuration.

    SLUG is the slug of the workspace to remove from the configuration.
    """
    user_config = open_config()
    if slug not in user_config["workspaces"]:
        click.echo(f"Workspace {slug} does not exist")
        return
    click.echo(f"Removing workspace {slug}")

    del user_config["workspaces"][slug]
    if f"pipelines.{slug}" in user_config.sections():
        del user_config[f"pipelines.{slug}"]

    if user_config["openhexa"]["current_workspace"] == slug:
        user_config["openhexa"].update({"current_workspace": ""})

    save_config(user_config)


@app.group(invoke_without_command=True)
@click.pass_context
def config(ctx):
    """
    Manage configuration of the CLI.
    """

    if ctx.invoked_subcommand is None:
        user_config = open_config()
        click.echo("Debug: " + ("True" if is_debug(user_config) else "False"))
        click.echo(f"Backend URL: {user_config['openhexa']['url']}")
        try:
            click.echo(
                f"Current workspace: {user_config['openhexa']['current_workspace']}"
            )
        except KeyError:
            click.echo("No current workspace")
        click.echo("\nWorkspaces:")
        click.echo("\n".join(user_config["workspaces"].keys()))


@config.command(name="set_url")
@click.argument("url")
def config_set_url(url):
    """
    Set the URL of the backend.

    """
    user_config = open_config()
    user_config["openhexa"].update({"url": url})
    save_config(user_config)
    click.echo(f"Set backend URL to {user_config['openhexa']['url']}")


@app.group(invoke_without_command=True)
@click.pass_context
def pipelines(ctx):
    """
    Manage pipelines (list pipelines, push a pipeline to the backend)
    """
    if ctx.invoked_subcommand is None:
        ctx.forward(pipelines_list)


@pipelines.command("init")
@click.argument("name", type=str)
def pipelines_init(name: str):
    """
    Initialize a new pipeline in a fresh directory.
    """

    new_pipeline_directory_name = stringcase.snakecase(name.lower())
    new_pipeline_path = Path.cwd() / Path(new_pipeline_directory_name)
    if new_pipeline_path.exists():
        click.echo(
            f"There is already a {name} directory in the current directory. Please choose a new name "
            f"for your pipeline.",
            err=True,
        )
        sys.exit(1)

    # Load samples
    sample_directory_path = Path(__file__).parent / Path("skeleton")
    with open(sample_directory_path / Path(".gitignore"), "r") as sample_ignore_file:
        sample_ignore_content = sample_ignore_file.read()
    with open(sample_directory_path / Path("pipeline.py"), "r") as sample_pipeline_file:
        sample_pipeline_content = (
            sample_pipeline_file.read()
            .replace("skeleton-pipeline-code", stringcase.spinalcase(name.lower()))
            .replace("skeleton_pipeline_name", stringcase.snakecase(name.lower()))
            .replace("Skeleton pipeline name", name)
        )
    with open(
        sample_directory_path / Path("workspace.yaml"), "r"
    ) as sample_workspace_file:
        sample_workspace_content = sample_workspace_file.read()

    # Create directory
    new_pipeline_path.mkdir(exist_ok=False)
    (new_pipeline_path / Path("workspace")).mkdir(exist_ok=False)
    with open(new_pipeline_path / Path(".gitignore"), "w") as ignore_file:
        ignore_file.write(sample_ignore_content)
    with open(new_pipeline_path / Path("pipeline.py"), "w") as pipeline_file:
        pipeline_file.write(sample_pipeline_content)
    with open(new_pipeline_path / Path("workspace.yaml"), "w") as workspace_file:
        workspace_file.write(sample_workspace_content)

    # Success
    click.echo(
        f"{click.style('Success!', fg='green')} Your pipeline has been created in "
        f"the {new_pipeline_directory_name}/ directory"
    )
    click.echo(
        f"Run it using {click.style(f'cd {new_pipeline_directory_name}', fg='cyan')} && "
        f"{click.style('python pipeline.py', fg='cyan')}"
    )


@pipelines.command("push")
@click.argument(
    "path", default=".", type=click.Path(exists=True, file_okay=False, dir_okay=True)
)
def pipelines_push(path: str):
    """
    Push a pipeline to the backend. If the pipeline already exists, it will be updated otherwise it will be created.

    PATH is the path to the pipeline file.
    """

    user_config = open_config()
    try:
        workspace = user_config["openhexa"]["current_workspace"]
    except KeyError:
        click.echo(
            "No workspace activated. Use openhexa workspaces add or openhexa workspaces activate to "
            "activate a workspace.",
            err=True,
        )
        sys.exit(1)

    ensure_is_pipeline_dir(path)

    try:
        pipeline = import_pipeline(path)
    except Exception as e:
        click.echo(f'Error while importing pipeline: "{e}"', err=True)
        if is_debug(user_config):
            raise e
        sys.exit(1)
    else:
        workspace_pipelines = get_pipelines(user_config)
        if is_debug(user_config):
            click.echo(workspace_pipelines)

        if get_pipeline(user_config, pipeline.code) is None:
            click.echo(
                f"Pipeline {click.style(pipeline.code, bold=True)} found in {path} does not exist in workspace {click.style(workspace, bold=True)}"
            )
            click.confirm(
                f"Create pipeline {click.style(pipeline.code, bold=True)} in workspace {click.style(workspace, bold=True)}?",
                True,
                abort=True,
            )
            create_pipeline(user_config, pipeline.code, pipeline.name)

        click.echo(
            f"Pushing pipeline {click.style(pipeline.code, bold=True)} to workspace {click.style(workspace, bold=True)}"
        )

        new_version = upload_pipeline(user_config, path)
        click.echo(f"New version created: {new_version}")

        url = f"{user_config['openhexa']['url']}/workspaces/{workspace}/pipelines/{pipeline.code}".replace(
            "api", "app"
        )
        click.echo(
            f"Done! You can view the pipeline in OpenHexa on {click.style(url, fg='bright_blue', underline=True)}"
        )


@pipelines.command("list")
def pipelines_list():
    """
    List all the remote pipelines of the current workspace.
    """
    user_config = open_config()
    workspace = user_config["openhexa"]["current_workspace"]

    if workspace == "":
        click.echo("No workspace activated", err=True)
        sys.exit(1)

    workspace_pipelines = get_pipelines(user_config)
    if len(workspace_pipelines) == 0:
        click.echo(f"No pipelines in workspace {workspace}")
        return
    click.echo("Pipelines:")
    for pipeline in workspace_pipelines:
        version = pipeline["currentVersion"].get("number")
        if version:
            version = f"v{version}"
        else:
            version = "N/A"
        click.echo(f"* {pipeline['code']} - {pipeline['name']} ({version})")



from __future__ import annotations

import argparse
import datetime
import json
import os
import string
import time
import typing
from logging import getLogger

import requests
from multiprocess import get_context  # NOQA

from openhexa.sdk.utils import Environments, get_environment

from .parameter import FunctionWithParameter, Parameter, ParameterValueError
from .task import PipelineWithTask
from .utils import load_local_workspace_config

logger = getLogger(__name__)


class PipelineConfigError(Exception):
    pass


def pipeline(
    code: str, *, name: str = None
) -> typing.Callable[[typing.Callable[..., typing.Any]], "Pipeline"]:
    if any(c not in string.ascii_lowercase + string.digits + "_-" for c in code):
        raise Exception(
            "Pipeline name should contains only lower case letters, digits, '_' and '-'"
        )

    def decorator(fun):
        if isinstance(fun, FunctionWithParameter):
            parameters = fun.get_all_parameters()
        else:
            parameters = []

        return Pipeline(code, name, fun, parameters)

    return decorator


class PipelineRunError(Exception):
    pass


class Pipeline:
    def __init__(
        self,
        code: str,
        name: str,
        function: typing.Callable,
        parameters: typing.Sequence[Parameter],
    ):
        self.code = code
        self.name = name
        self.function = function
        self.parameters = parameters
        self.tasks = []

    def task(self, function):
        """task decorator"""

        return PipelineWithTask(function, self)

    def run(self, config: typing.Dict[str, typing.Any]):
        now = datetime.datetime.utcnow().replace(microsecond=0).isoformat()
        print(f'{now} Starting pipeline "{self.code}"')

        # Validate / default parameters
        validated_config = {}
        for parameter in self.parameters:
            value = config.pop(parameter.code, None)
            validated_value = parameter.validate(value)
            validated_config[parameter.code] = validated_value

        if len(config) > 0:
            raise ParameterValueError(
                f"The provided config contains invalid key(s): {', '.join(list(config.keys()))}"
            )

        self.function(**validated_config)

        # managing variables
        result_list = []
        context = get_context("spawn")
        pool = context.Pool()  # FIXME: set max size of pool
        total = len(self.tasks)
        completed = 0

        while True:
            tasks = self.get_available_tasks()
            # filter already pooled task, even if they are not finished
            tasks = [t for t in tasks if not t.pooled]

            if len(tasks) == 0 and len(result_list) == 0:
                # nothing running, nothing to do
                break

            for task in tasks:
                now = datetime.datetime.utcnow().replace(microsecond=0).isoformat()
                print(f'{now} Started task "{task.compute.__name__}"')
                result = pool.apply_async(task.run)
                result_list.append((result, task))
                task.pooled = True

            while True:
                dag_step = False

                for result, task in result_list:
                    if not result.ready():
                        continue
                    now = (
                        datetime.datetime.now(tz=datetime.timezone.utc)
                        .replace(microsecond=0)
                        .isoformat()
                    )

                    completed += 1
                    progress = int(completed / total * 100)
                    print(f'{now} Finished task "{task.compute.__name__}"')
                    self._update_progress(progress)

                    try:
                        task_com_result = result.get()
                        task.result = task_com_result.result
                        task.start_time = task_com_result.start_time
                        task.end_time = task_com_result.end_time
                        dag_step = True
                    except Exception as e:  # NOQA
                        raise PipelineRunError(f"Pipeline {self.code} failed: {e}")

                if dag_step:
                    # remove finished tasks
                    result_list = [(r, t) for r, t in result_list if t.end_time is None]
                    break
                else:
                    # busy loop
                    time.sleep(0.3)

        pool.close()
        pool.join()

        print(f'{now} Successfully completed pipeline "{self.code}"')

    def get_available_tasks(self):
        return [task for task in self.tasks if task.is_ready()]

    def parameters_spec(self):
        return [arg.parameter_spec() for arg in self.parameters]

    def _update_progress(self, progress: int):
        if self.connected:
            token = os.environ["HEXA_TOKEN"]
            headers = {"Authorization": "Bearer %s" % token}
            query = """
                            mutation updatePipelineProgress ($input: UpdatePipelineProgressInput!) {
                                updatePipelineProgress(input: $input) { success errors }
                            }"""
            r = requests.post(
                f'{os.environ["HEXA_SERVER_URL"]}/graphql/',
                headers=headers,
                json={
                    "query": query,
                    "variables": {"input": {"percent": progress}},
                },
            )
            r.raise_for_status()
        else:
            print(f"Progress update: {progress}%")

    @property
    def connected(self):
        env = get_environment()
        return env == Environments.PIPELINE and "HEXA_SERVER_URL" in os.environ

    def __call__(self, config: typing.Optional[typing.Dict[str, typing.Any]] = None):
        # Handle config
        if config is None:  # Called without arguments, in the pipeline file itself
            parser = argparse.ArgumentParser(exit_on_error=False)
            parser.add_argument("-c", "--config")
            parser.add_argument("-f", "--config-file")
            # We can't use parse_args, as it will call sys.exit() if there are unrecognized arguments
            args, argv = parser.parse_known_args()
            if argv or (args.config_file is not None and args.config is not None):
                raise ValueError(
                    f"Unrecognized arguments: {' '.join(argv)}. Running a pipeline requires a single "
                    "argument: either an inline JSON config with the --values/-v argument, or a JSON "
                    "config file with the --values-file/-f argument."
                )
            if args.config_file is not None:
                with open(args.config_file, "r") as cf:
                    try:
                        config = json.load(cf)
                    except json.JSONDecodeError:
                        raise PipelineConfigError(
                            "The provided config is not valid JSON"
                        )

            elif args.config is not None:
                try:
                    config = json.loads(args.config)
                except json.JSONDecodeError:
                    raise PipelineConfigError("The provided config is not valid JSON")
            else:
                config = {}

        # Handle local workspace config for dev / testing, if appropriate
        if not self.connected:
            load_local_workspace_config()

        self.run(config)



from unittest.mock import Mock

import pytest

from openhexa.sdk.pipelines.parameter import Parameter, ParameterValueError
from openhexa.sdk.pipelines.pipeline import Pipeline


def test_pipeline_run_valid_config():
    pipeline_func = Mock()
    parameter_1 = Parameter("arg1", type=str)
    parameter_2 = Parameter("arg2", type=str, multiple=True)
    parameter_3 = Parameter("arg3", type=int, default=33)
    pipeline = Pipeline(
        "code", "pipeline", pipeline_func, [parameter_1, parameter_2, parameter_3]
    )
    pipeline.run({"arg1": "ab", "arg2": ["cd", "ef"]})

    assert pipeline.name == "pipeline"
    pipeline_func.assert_called_once_with(arg1="ab", arg2=["cd", "ef"], arg3=33)


def test_pipeline_run_invalid_config():
    pipeline_func = Mock()
    parameter_1 = Parameter("arg1", type=str)
    pipeline = Pipeline("code", "pipeline", pipeline_func, [parameter_1])
    with pytest.raises(ParameterValueError):
        pipeline.run({"arg1": 3})


def test_pipeline_run_extra_config():
    pipeline_func = Mock()
    parameter_1 = Parameter("arg1", type=str)
    pipeline = Pipeline("code", "pipeline", pipeline_func, [parameter_1])
    with pytest.raises(ParameterValueError):
        pipeline.run({"arg1": "ok", "arg2": "extra"})


def test_pipeline_parameters_spec():
    pipeline_func = Mock()
    parameter_1 = Parameter("arg1", type=str)
    parameter_2 = Parameter("arg2", type=str, multiple=True)
    pipeline = Pipeline("code", "pipeline", pipeline_func, [parameter_1, parameter_2])

    assert pipeline.parameters_spec() == [
        {
            "code": "arg1",
            "name": None,
            "type": "str",
            "required": True,
            "choices": None,
            "help": None,
            "multiple": False,
            "default": None,
        },
        {
            "code": "arg2",
            "name": None,
            "type": "str",
            "required": True,
            "choices": None,
            "help": None,
            "multiple": True,
            "default": None,
        },
    ]



import datetime
import os
import typing

import requests

from openhexa.sdk.utils import Environments, get_environment
from openhexa.sdk.workspaces import workspace


class CurrentRun:
    @property
    def connected(self):
        env = get_environment()
        return env == Environments.PIPELINE and "HEXA_SERVER_URL" in os.environ

    def add_file_output(self, path: str):
        stripped_path = path.replace(workspace.files_path, "")
        name = stripped_path.strip("/")
        if self.connected:
            query = """
                mutation addPipelineOutput ($input: AddPipelineOutputInput!) {
                    addPipelineOutput(input: $input) { success errors }
                }"""
            variables = {
                "input": {
                    "uri": f"gs://{os.environ['WORKSPACE_BUCKET_NAME']}{stripped_path}",
                    "type": "file",
                    "name": name,
                }
            }
            self._graphql_query(query, variables)
        else:
            print(f"Sending output with path {stripped_path}")

    def add_database_output(self, table_name: str):
        if self.connected:
            query = """
                mutation addPipelineOutput ($input: AddPipelineOutputInput!) {
                    addPipelineOutput(input: $input) { success errors }
                }"""
            variables = {
                "input": {
                    "uri": f"postgresql://{workspace.database_host}/{workspace.database_name}/{table_name}",
                    "type": "db",
                    "name": table_name,
                }
            }
            self._graphql_query(query, variables)
        else:
            print(f"Sending output with table_name {table_name}")

    def log_debug(self, message: str):
        self._log_message("DEBUG", message)

    def log_info(self, message: str):
        self._log_message("INFO", message)

    def log_warning(self, message: str):
        self._log_message("WARNING", message)

    def log_error(self, message: str):
        self._log_message("ERROR", message)

    def log_critical(self, message: str):
        self._log_message("CRITICAL", message)

    def _log_message(
        self,
        priority: typing.Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        message: str,
    ):
        valid_priorities = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
        if priority not in valid_priorities:
            raise ValueError(f"priority must be one of {', '.join(valid_priorities)}")

        if self.connected:
            query = """
                    mutation logPipelineMessage ($input: LogPipelineMessageInput!) {
                        logPipelineMessage(input: $input) { success errors }
                    }"""
            variables = {"input": {"priority": priority, "message": message}}
            self._graphql_query(query, variables)
        else:
            now = (
                datetime.datetime.now(tz=datetime.timezone.utc)
                .replace(microsecond=0)
                .isoformat()
            )
            print(now, priority, message)

    @staticmethod
    def _graphql_query(
        query: str, variables: typing.Optional[typing.Dict[str, typing.Any]] = None
    ):
        token = os.environ["HEXA_TOKEN"]
        headers = {"Authorization": "Bearer %s" % token}
        r = requests.post(
            f'{os.environ["HEXA_SERVER_URL"]}/graphql/',
            headers=headers,
            json={
                "query": query,
                "variables": variables if variables is not None else {},
            },
        )
        r.raise_for_status()
        return r.json()


current_run = CurrentRun()
